-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local Auctioning = TSM.Operations:NewPackage("Auctioning")
local ClientInfo = TSM.LibTSMWoW:Include("Util.ClientInfo")
local L = TSM.Include("Locale").GetTable()
local TempTable = TSM.LibTSMUtil:Include("BaseType.TempTable")
local Money = TSM.LibTSMUtil:Include("UI.Money")
local Operation = TSM.LibTSMTypes:Include("Operation")
local Group = TSM.LibTSMTypes:Include("Group")
local private = {}
local OPERATION_VALUE_LIMITS = {
	postCap = { min = 0, max = 50000 },
	keepQuantity = { min = 0, max = 50000 },
	maxExpires = { min = 0, max = 50000 },
}
if ClientInfo.HasFeature(ClientInfo.FEATURES.AH_STACKS) then
	OPERATION_VALUE_LIMITS.stackSize = { min = 1, max = 200 }
	OPERATION_VALUE_LIMITS.postCap.max = 200
end



-- ============================================================================
-- Module Functions
-- ============================================================================

function Auctioning.OnInitialize()
	local operationType = Operation.NewType("Auctioning", L["Auctioning"], 20)
		:SetDescriptionFunc(private.GetDescription)
		:SetCustomSanitizeFunc(private.SanitizeSettings)
		:AddNumberSetting("ignoreLowDuration")
		:AddStringSetting("postCap", "5")
		:AddStringSetting("keepQuantity", "0")
		:AddStringSetting("maxExpires", "0")
		:AddNumberSetting("duration", 2, private.SanitizeDuration)
		:AddNumberSetting("bidPercent", 1)
		:AddStringSetting("undercut", ClientInfo.IsRetail() and "0c" or "1c", private.SanitizeUndercut)
		:AddStringSetting("minPrice", "check(first(crafting,dbmarket,dbregionmarketavg),max(0.25*avg(crafting,dbmarket,dbregionmarketavg),1.5*vendorsell))")
		:AddStringSetting("maxPrice", "check(first(crafting,dbmarket,dbregionmarketavg),max(5*avg(crafting,dbmarket,dbregionmarketavg),30*vendorsell))")
		:AddStringSetting("normalPrice", "check(first(crafting,dbmarket,dbregionmarketavg),max(2*avg(crafting,dbmarket,dbregionmarketavg),12*vendorsell))")
		:AddStringSetting("priceReset", "none")
		:AddStringSetting("aboveMax", "maxPrice")
		:AddBooleanSetting("cancelUndercut", true)
		:AddBooleanSetting("cancelRepost", true)
		:AddStringSetting("cancelRepostThreshold", "1g")
	if not ClientInfo.IsRetail() then
		operationType:AddStringSetting("blacklist")
	end
	if ClientInfo.HasFeature(ClientInfo.FEATURES.AH_STACKS) then
		operationType:AddBooleanSetting("matchStackSize")
		operationType:AddStringSetting("stackSize", "1")
		operationType:AddBooleanSetting("stackSizeIsCap")
	end
	TSM.Operations.RegisterType(operationType)
end

function Auctioning.GetMinMaxValues(key)
	local info = OPERATION_VALUE_LIMITS[key]
	return info and info.min or -math.huge, info and info.max or math.huge
end

function Auctioning.GetMinPrice(itemString)
	return private.GetOperationValueHelper(itemString, "minPrice")
end

function Auctioning.GetMaxPrice(itemString)
	return private.GetOperationValueHelper(itemString, "maxPrice")
end

function Auctioning.GetNormalPrice(itemString)
	return private.GetOperationValueHelper(itemString, "normalPrice")
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.SanitizeSettings(operation)
	if not ClientInfo.HasFeature(ClientInfo.FEATURES.AH_STACKS) and operation.stackSize then
		operation.postCap = tonumber(operation.postCap) * tonumber(operation.stackSize)
	end
	if ClientInfo.IsRetail() and (type(operation.undercut) == "number" and operation.undercut or Money.FromString(operation.undercut) or math.huge) < COPPER_PER_SILVER then
		operation.undercut = "0c"
	end
end

function private.SanitizeDuration(value)
	-- convert from 12/24/48 durations to 1/2/3 API values
	if value == 12 then
		return 1
	elseif value == 24 then
		return 2
	elseif value == 48 then
		return 3
	else
		return value
	end
end

function private.SanitizeUndercut(value)
	if ClientInfo.IsRetail() and (Money.FromString(Money.ToStringExact(value) or value) or math.huge) < COPPER_PER_SILVER then
		return "0c"
	end
	return value
end

function private.GetOperationValueHelper(itemString, key)
	local origItemString = itemString
	itemString = Group.TranslateItemString(itemString)
	local operationName, operationSettings = TSM.Operations.GetFirstOperationByItem("Auctioning", itemString)
	if not operationName then
		return
	end
	return TSM.Auctioning.Util.GetPrice(key, operationSettings, origItemString)
end

function private.GetDescription(operationSettings)
	local parts = TempTable.Acquire()

	-- get the post string
	if operationSettings.postCap == 0 then
		tinsert(parts, L["No posting."])
	else
		if ClientInfo.HasFeature(ClientInfo.FEATURES.AH_STACKS) then
			tinsert(parts, format(L["Posting %d stack(s) of %d for %s hours."], operationSettings.postCap, operationSettings.stackSize, strmatch(TSM.CONST.AUCTION_DURATIONS[operationSettings.duration], "%d+")))
		else
			tinsert(parts, format(L["Posting %d items for %s hours."], operationSettings.postCap, strmatch(TSM.CONST.AUCTION_DURATIONS[operationSettings.duration], "%d+")))
		end
	end

	-- get the cancel string
	if operationSettings.cancelUndercut and operationSettings.cancelRepost then
		tinsert(parts, format(L["Canceling undercut auctions and to repost higher."]))
	elseif operationSettings.cancelUndercut then
		tinsert(parts, format(L["Canceling undercut auctions."]))
	elseif operationSettings.cancelRepost then
		tinsert(parts, format(L["Canceling to repost higher."]))
	else
		tinsert(parts, L["Not canceling."])
	end

	local result = table.concat(parts, " ")
	TempTable.Release(parts)
	return result
end
