-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local ChatMessage = TSM.Include("Util.ChatMessage")
local Log = TSM.LibTSMUtil:Include("Util.Log")
local MoneyFormatter = TSM.LibTSMUtil:IncludeClassType("MoneyFormatter")
local StaticPopupDialog = TSM.LibTSMWoW:IncludeClassType("StaticPopupDialog")
local Addon = TSM.LibTSMWoW:Include("Util.Addon")
local ScriptWrapper = TSM.LibTSMWoW:Include("API.ScriptWrapper")
local Event = TSM.LibTSMWoW:Include("Service.Event")
local DelayTimer = TSM.LibTSMWoW:IncludeClassType("DelayTimer")
local CustomString = TSM.LibTSMTypes:Include("CustomString")
local ItemInfo = TSM.LibTSMService:Include("Item.ItemInfo")
local ItemFilter = TSM.LibTSMService:IncludeClassType("ItemFilter")
local VendorBuy = TSM.LibTSMService:Include("Item.VendorBuy")
local Inventory = TSM.Include("Service.Inventory")
local L = TSM.Include("Locale").GetTable()
local private = {
	settings = nil,
	itemInfoPublisher = nil,  --luacheck: ignore 1004 - just stored for GC reasons
	oribosExchangeTemp = {},
}
local SCRIPT_CALLBACK_TIME_WARNING_THRESHOLD = 0.02
local DELAY_CALLBACK_TIME_WARNING_THRESHOLD = 0.05
local EVENT_CALLBACK_TIME_WARNING_THRESHOLD = 0.02
do
	-- Show a message if we were updated
	if GetAddOnMetadata("TradeSkillMaster", "Version") ~= "v4.13.15" and not TSM.IsTest() then
		StaticPopupDialog.ShowWithOk("TSM was just updated and may not work properly until you restart WoW.")
	end
	-- Basic module configuration
	Log.SetLoggingToChatEnabled(TSM.IsTest())
	MoneyFormatter.SetLargeNumberSeperator(LARGE_NUMBER_SEPERATOR)
end



-- ============================================================================
-- Module Functions
-- ============================================================================

---@param settingsDB SettingsDB
function TSM.OnInitialize(settingsDB)
	-- Load settings
	private.settings = settingsDB:NewView()
		:AddKey("global", "coreOptions", "chatFrame")
		:AddKey("global", "coreOptions", "destroyValueSource")
		:AddKey("global", "debug", "chatLoggingEnabled")
		:AddKey("sync", "internalData", "classKey")
		:RegisterCallback("destroyValueSource", function() CustomString.InvalidateCache("Destroy") end)

	-- Configure some LibTSM modules
	ScriptWrapper.SetTimeThreshold(SCRIPT_CALLBACK_TIME_WARNING_THRESHOLD, function(script, objStr, elapsed)
		Log.Warn("Script handler (%s) for frame (%s) took %0.5fs", script, objStr, elapsed)
	end)
	DelayTimer.SetTimeThreshold(DELAY_CALLBACK_TIME_WARNING_THRESHOLD, function(name, elapsed)
		Log.Warn("Delay callback (%s) took %0.5fs", name, elapsed)
	end)
	Event.SetTimeThreshold(EVENT_CALLBACK_TIME_WARNING_THRESHOLD, function(event, elapsed)
		Log.Warn("Event (%s) callback took %0.5fs", event, elapsed)
	end)

	-- Configure the logger
	ChatMessage.SetChatFrame(private.settings.chatFrame)
	if TSM.IsTest() then
		private.settings.chatLoggingEnabled = true
	elseif not TSM.IsDev() then
		private.settings.chatLoggingEnabled = false
	end
	Log.SetLoggingToChatEnabled(private.settings.chatLoggingEnabled)

	-- Store the class of this character
	private.settings.classKey = select(2, UnitClass("player"))

	-- CanIMogIt integration
	-- luacheck: globals CanIMogIt
	if CanIMogIt and CanIMogIt.PlayerKnowsTransmog then
		ItemFilter.AddKeyMatchPart("unlearned", function(item)
			return CanIMogIt:PlayerKnowsTransmog(ItemInfo.GetLink(item))
		end)
	end
	if CanIMogIt and CanIMogIt.CharacterCanLearnTransmog then
		ItemFilter.AddKeyMatchPart("canlearn", function(item)
			return CanIMogIt:CharacterCanLearnTransmog(ItemInfo.GetLink(item))
		end)
	end

	-- Core price sources
	private.itemInfoPublisher = ItemInfo.GetPublisher()
		:CallFunction(function(itemString)
			CustomString.InvalidateCache("VendorBuy", itemString)
			CustomString.InvalidateCache("VendorSell", itemString)
			CustomString.InvalidateCache("ItemQuality", itemString)
			CustomString.InvalidateCache("ItemLevel", itemString)
			CustomString.InvalidateCache("RequiredLevel", itemString)
		end)
		:Stored()
	CustomString.RegisterSource("Item", "VendorBuy", L["Buy from Vendor"], VendorBuy.Get, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Item", "VendorSell", L["Sell to Vendor"], ItemInfo.GetVendorSell, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Item", "ItemQuality", L["Item Quality"], ItemInfo.GetQuality, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Item", "ItemLevel", L["Item Level"], ItemInfo.GetItemLevel, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Item", "RequiredLevel", L["Required Level"], ItemInfo.GetMinLevel, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Item", "NumInventory", L["Total Inventory Quantity"], Inventory.GetTotalQuantity, CustomString.SOURCE_TYPE.NORMAL)

	-- Auctioneer price sources
	-- luacheck: globals AucAdvanced
	if Addon.IsEnabled("Auc-Advanced") and AucAdvanced then
		local registeredAuctioneerSources = {}
		hooksecurefunc(AucAdvanced, "SendProcessorMessage", function(msg)
			if msg == "scanfinish" then
				for _, source in ipairs(registeredAuctioneerSources) do
					CustomString.InvalidateCache(source)
				end
			end
		end)
		if AucAdvanced.Modules.Util.Appraiser and AucAdvanced.Modules.Util.Appraiser.GetPrice then
			local function PriceFunc(itemString)
				local itemLink = ItemInfo.GetLink(itemString)
				if not itemLink then
					return nil
				end
				return AucAdvanced.Modules.Util.Appraiser.GetPrice(itemLink)
			end
			CustomString.RegisterSource("External", "AucAppraiser", L["Auctioneer - Appraiser"], PriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
			tinsert(registeredAuctioneerSources, "AucAppraiser")
		end
		if AucAdvanced.Modules.Util.SimpleAuction and AucAdvanced.Modules.Util.SimpleAuction.Private.GetItems then
			local function PriceFunc(itemString)
				local itemLink = ItemInfo.GetLink(itemString)
				if not itemLink then
					return nil
				end
				return select(6, AucAdvanced.Modules.Util.SimpleAuction.Private.GetItems(itemLink)) or nil
			end
			CustomString.RegisterSource("External", "AucMinBuyout", L["Auctioneer - Minimum Buyout"], PriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
			tinsert(registeredAuctioneerSources, "AucMinBuyout")
		end
		if AucAdvanced.API.GetMarketValue then
			local function PriceFunc(itemString)
				local itemLink = ItemInfo.GetLink(itemString)
				if not itemLink then
					return nil
				end
				return AucAdvanced.API.GetMarketValue(itemLink)
			end
			CustomString.RegisterSource("External", "AucMarket", L["Auctioneer - Market Value"], PriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
			tinsert(registeredAuctioneerSources, "AucMarket")
		end
	end

	-- Auctionator price sources
	if Addon.IsEnabled("Auctionator") then
		local PriceFunc = nil
		-- luacheck: globals Auctionator Atr_GetAuctionBuyout Atr_RegisterFor_DBupdated
		if Auctionator and Auctionator.API and Auctionator.API.v1 and Auctionator.API.v1.RegisterForDBUpdate then
			-- Retail version
			local ok = pcall(function()
				Auctionator.API.v1.RegisterForDBUpdate("TradeSkillMaster", function() CustomString.InvalidateCache("AtrValue") end)
			end)
			if ok then
				PriceFunc = function(itemString)
					local itemLink = ItemInfo.GetLink(itemString)
					if not itemLink then
						return nil
					end
					return Auctionator.API.v1.GetAuctionPriceByItemLink("TradeSkillMaster", itemLink)
				end
			end
		elseif Atr_GetAuctionBuyout and Atr_RegisterFor_DBupdated then
			-- Classic version
			Atr_RegisterFor_DBupdated(function()
				CustomString.InvalidateCache("AtrValue")
			end)
			PriceFunc = function(itemString)
				local itemLink = ItemInfo.GetLink(itemString)
				if not itemLink then
					return nil
				end
				return Atr_GetAuctionBuyout(itemLink)
			end
		end
		if PriceFunc then
			CustomString.RegisterSource("External", "AtrValue", L["Auctionator - Auction Value"], PriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
		end
	end

	-- OribosExchange price sources
	-- luacheck: globals OEMarketInfo
	if Addon.IsEnabled("OribosExchange") and OEMarketInfo then
		local function PriceFuncHelper(itemString, key)
			local itemLink = ItemInfo.GetLink(itemString)
			if not itemLink then
				return nil
			end
			local data = OEMarketInfo(itemLink, private.oribosExchangeTemp)
			return data and data[key] or nil
		end
		local function RealmPriceFunc(itemString)
			return PriceFuncHelper(itemString, "realm")
		end
		CustomString.RegisterSource("External", "OERealm", L["Oribos Exchange Realm Price"], RealmPriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
		local function RegionPriceFunc(itemString)
			return PriceFuncHelper(itemString, "region")
		end
		CustomString.RegisterSource("External", "OERegion", L["Oribos Exchange Region Price"], RegionPriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
	end

	-- AHDB price sources
	-- luacheck: globals AuctionDB
	if Addon.IsEnabled("AuctionDB") and AuctionDB and AuctionDB.AHGetAuctionInfoByLink then
		hooksecurefunc(AuctionDB, "AHendOfScanCB", function()
			CustomString.InvalidateCache("AHDBMinBuyout")
			CustomString.InvalidateCache("AHDBMinBid")
		end)
		local function PriceFuncHelper(itemString, key)
			local itemLink = ItemInfo.GetLink(itemString)
			if not itemLink then
				return nil
			end
			local info = AuctionDB:AHGetAuctionInfoByLink(itemLink)
			return info and info[key] or nil
		end
		local function MinBuyoutPriceFunc(itemString)
			return PriceFuncHelper(itemString, "minBuyout")
		end
		CustomString.RegisterSource("External", "AHDBMinBuyout", L["AHDB Minimum Buyout"], MinBuyoutPriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
		local function MinBidPriceFunc(itemString)
			return PriceFuncHelper(itemString, "minBid")
		end
		CustomString.RegisterSource("External", "AHDBMinBid", L["AHDB Minimum Bid"], MinBidPriceFunc, CustomString.SOURCE_TYPE.PRICE_DB)
	end

	-- Accounting sources
	CustomString.RegisterSource("Accounting", "AvgSell", L["Avg Sell Price"], TSM.Accounting.Transactions.GetAverageSalePrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "MaxSell", L["Max Sell Price"], TSM.Accounting.Transactions.GetMaxSalePrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "MinSell", L["Min Sell Price"], TSM.Accounting.Transactions.GetMinSalePrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "AvgBuy", L["Avg Buy Price"], TSM.Accounting.Transactions.GetAverageBuyPrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "SmartAvgBuy", L["Smart Avg Buy Price"], TSM.Accounting.Transactions.GetSmartAverageBuyPrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "MaxBuy", L["Max Buy Price"], TSM.Accounting.Transactions.GetMaxBuyPrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "MinBuy", L["Min Buy Price"], TSM.Accounting.Transactions.GetMinBuyPrice, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "NumExpires", L["Expires Since Last Sale"], TSM.Accounting.Auctions.GetNumExpiresSinceSale, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Accounting", "SaleRate", L["Sale Rate"], TSM.Accounting.GetSaleRate, CustomString.SOURCE_TYPE.NORMAL)

	-- AuctionDB sources
	local function GetAuctionDBPriceFunc(key, isRegion)
		return function(itemString)
			if isRegion then
				return TSM.AuctionDB.GetRegionItemData(itemString, key)
			else
				return TSM.AuctionDB.GetRealmItemData(itemString, key)
			end
		end
	end
	CustomString.RegisterSource("AuctionDB", "DBMarket", L["AuctionDB - Market Value"], GetAuctionDBPriceFunc("marketValue"), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBMinBuyout", L["AuctionDB - Minimum Buyout"], GetAuctionDBPriceFunc("minBuyout"), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBRecent", L["AuctionDB - Recent Value"], GetAuctionDBPriceFunc("marketValueRecent"), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBHistorical", L["AuctionDB - Historical Price"], GetAuctionDBPriceFunc("historical"), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBRegionMarketAvg", L["AuctionDB - Region Market Value Average"], GetAuctionDBPriceFunc("regionMarketValue", true), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBRegionHistorical", L["AuctionDB - Region Historical Price"], GetAuctionDBPriceFunc("regionHistorical", true), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBRegionSaleAvg", L["AuctionDB - Region Sale Average"], GetAuctionDBPriceFunc("regionSale", true), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBRegionSaleRate", L["AuctionDB - Region Sale Rate"], GetAuctionDBPriceFunc("regionSalePercent", true), CustomString.SOURCE_TYPE.PRICE_DB)
	CustomString.RegisterSource("AuctionDB", "DBRegionSoldPerDay", L["AuctionDB - Region Sold Per Day"], GetAuctionDBPriceFunc("regionSoldPerDay", true), CustomString.SOURCE_TYPE.PRICE_DB)

	-- Crafting sources
	local function GetDestroyValue(itemString)
		return TSM.Crafting.GetConversionsValue(itemString, private.settings.destroyValueSource)
	end
	CustomString.RegisterSource("Crafting", "Destroy", L["Destroy Value"], GetDestroyValue, CustomString.SOURCE_TYPE.NORMAL)
	CustomString.RegisterSource("Crafting", "Crafting", L["Crafting Cost"], TSM.Crafting.Cost.GetLowestCostByItem, CustomString.SOURCE_TYPE.VOLATILE)
	CustomString.RegisterSource("Crafting", "MatPrice", L["Crafting Material Cost"], TSM.Crafting.Cost.GetMatCost, CustomString.SOURCE_TYPE.VOLATILE)

	-- Operation-based price sources
	CustomString.RegisterSource("Operations", "AuctioningOpMin", L["First Auctioning Operation Min Price"], TSM.Operations.Auctioning.GetMinPrice, CustomString.SOURCE_TYPE.VOLATILE)
	CustomString.RegisterSource("Operations", "AuctioningOpMax", L["First Auctioning Operation Max Price"], TSM.Operations.Auctioning.GetMaxPrice, CustomString.SOURCE_TYPE.VOLATILE)
	CustomString.RegisterSource("Operations", "AuctioningOpNormal", L["First Auctioning Operation Normal Price"], TSM.Operations.Auctioning.GetNormalPrice, CustomString.SOURCE_TYPE.VOLATILE)
	CustomString.RegisterSource("Operations", "ShoppingOpMax", L["Shopping Operation Max Price"], TSM.Operations.Shopping.GetMaxPrice, CustomString.SOURCE_TYPE.VOLATILE)
	CustomString.RegisterSource("Operations", "SniperOpMax", L["Sniper Operation Below Price"], TSM.Operations.Sniper.GetBelowPrice, CustomString.SOURCE_TYPE.VOLATILE)

	-- Force a garbage collection
	collectgarbage()
end
